# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Generate test configuration files, run tests."""

from __future__ import annotations

import dataclasses
import sys
import typing

import click

from . import defs
from . import parse
from . import util
from .run import base
from .run import req
from .run import tox
from .run import uv


if typing.TYPE_CHECKING:
    from typing import Final


PY_SUPPORTED: Final = "supported"
"""The magic Python version meaning "run for all supported Python versions"."""


@dataclasses.dataclass
class ConfigHolder:
    """Hold a `Config` object."""

    cfg: defs.Config | None = None
    """The `Config` object stashed by the main function."""


def extract_cfg(ctx: click.Context) -> defs.Config:
    """Extract the `Config` object that the main function built."""
    cfg_hold: Final = ctx.find_object(ConfigHolder)
    if cfg_hold is None:
        sys.exit("Internal error: no click config holder object")

    cfg: Final = cfg_hold.cfg
    if cfg is None:
        sys.exit("Internal error: no config in the click config holder")

    return cfg


def arg_features(_ctx: click.Context, _self: click.Parameter, value: bool) -> bool:  # noqa: FBT001
    """Display program features information and exit."""
    if not value:
        return value

    print("Features: " + " ".join(f"{name}={value}" for name, value in defs.FEATURES.items()))
    sys.exit(0)


@click.command(name="envs")
@click.pass_context
def cmd_list_envs(ctx: click.Context) -> None:
    """List the defined test environments."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)
        print("\n".join(name for name in sorted(proj.env)))
    except defs.Error as err:
        sys.exit(f"Could not list the test environments: {err}")


@click.command(name="pythons")
@click.pass_context
def cmd_list_pythons(ctx: click.Context) -> None:
    """List the supported Python versions."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)
        print("\n".join(proj.python_supported))
    except defs.Error as err:
        sys.exit(f"Could not list the supported Python versions: {err}")


@click.command(name="reqs")
@click.pass_context
def cmd_list_reqs(ctx: click.Context) -> None:
    """List the defined dependency groups."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)
        print("\n".join(sorted(proj.python_dependency_groups)))
    except defs.Error as err:
        sys.exit(f"Could not list the project dependency groups: {err}")


@click.group(name="list")
@click.pass_context
def cmd_list(ctx: click.Context) -> None:
    """List various configuration settings."""


cmd_list.add_command(cmd_list_envs)
cmd_list.add_command(cmd_list_pythons)
cmd_list.add_command(cmd_list_reqs)


@click.command(name="generate")
@click.option("--check", is_flag=True, help="check whether the file is up to date")
@click.option(
    "--group",
    "-g",
    type=str,
    required=True,
    help="the dependency group to generate the requirements file for",
)
@click.option("--diff", is_flag=True, help="in check mode, display a diff")
@click.option("--force", "-f", is_flag=True, help="only write the file out if it has changed")
@click.option("--noop", "-N", is_flag=True, help="no-operation mode; display what would be done")
@click.option(
    "--output",
    "-o",
    type=str,
    help="the name of the file to write to, or '-' for standard output",
)
@click.pass_context
def cmd_req_generate(  # noqa: PLR0913  # this function does many things
    ctx: click.Context,
    *,
    check: bool,
    diff: bool,
    force: bool,
    group: str,
    noop: bool,
    output: str | None,
) -> None:
    """Generate the `requirements/groupname.txt` file."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)

        runner = req.ReqRunner(cfg=cfg, proj=proj, posargs=[], group=group)
        runner.generate(check=check, diff=diff, force=force, noop=noop, output=output)
    except base.NeedsRegeneratingError as err:
        sys.exit(str(err))
    except defs.Error as err:
        sys.exit(f"Could not generate the requirements file: {err}")


@click.group(name="req")
@click.pass_context
def cmd_req(ctx: click.Context) -> None:
    """Export dependency groups as requirement text files."""


cmd_req.add_command(cmd_req_generate)


@click.command(name="generate")
@click.option("--check", is_flag=True, help="check whether the file is up to date")
@click.option("--diff", is_flag=True, help="in check mode, display a diff")
@click.option("--force", "-f", is_flag=True, help="only write the file out if it has changed")
@click.option("--noop", "-N", is_flag=True, help="no-operation mode; display what would be done")
@click.option(
    "--output",
    "-o",
    type=str,
    default="tox.ini",
    help="the name of the file to write to, or '-' for standard output",
)
@click.pass_context
def cmd_tox_generate(  # noqa: PLR0913  # this function does many things
    ctx: click.Context,
    *,
    check: bool,
    diff: bool,
    force: bool,
    noop: bool,
    output: str,
) -> None:
    """Generate the `tox.ini` file."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)

        if cfg.python is None:
            cfg.log.info("Generating tox.ini using the default Python version")
        else:
            cfg.log.info("Generating tox.ini using  Python %(pyver)s", {"pyver": cfg.python})

        runner = tox.ToxRunner(cfg=cfg, proj=proj, posargs=[])
        runner.generate(check=check, diff=diff, force=force, noop=noop, output=output)
    except base.NeedsRegeneratingError as err:
        sys.exit(str(err))
    except defs.Error as err:
        sys.exit(f"Could not generate the Tox configuration: {err}")


@click.command(name="run")
@click.option("--env", "-e", type=str, help="the environments to run commands for")
@click.option("--force", "-f", is_flag=True, help="always write the tox.ini file out")
@click.option("--noop", "-N", is_flag=True, help="no-operation mode; display what would be done")
@click.argument("posargs", type=str, required=False, nargs=-1)
@click.pass_context
def cmd_tox_run(
    ctx: click.Context,
    *,
    env: str,
    force: bool,
    noop: bool,
    posargs: tuple[str],
) -> None:
    """Run test environments using Tox."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)
        posargs_list: Final = list(posargs)

        py_versions: Final = [cfg.python] if cfg.python != PY_SUPPORTED else proj.python_supported
        for pyver in py_versions:
            if pyver is None:
                cfg.log.info("Running the tests for the default Python version")
            else:
                cfg.log.info("Running the tests for Python %(pyver)s", {"pyver": pyver})

            runner = tox.ToxRunner(
                cfg=dataclasses.replace(cfg, python=pyver),
                proj=proj,
                posargs=posargs_list,
            )
            runner.generate(check=False, diff=False, force=force, noop=noop)
            runner.run(envs=env.split(",") if env is not None else [], noop=noop)
    except defs.Error as err:
        sys.exit(f"Could not run the tests using Tox: {err}")


@click.group(name="tox")
@click.pass_context
def cmd_tox(ctx: click.Context) -> None:
    """Generate the `tox.ini` file and run tests using Tox."""


cmd_tox.add_command(cmd_tox_generate)
cmd_tox.add_command(cmd_tox_run)


@click.command(name="run")
@click.option("--env", "-e", type=str, help="the environments to run commands for")
@click.option("--force", "-f", is_flag=True, help="unused for `uv`")
@click.option("--noop", "-N", is_flag=True, help="no-operation mode; display what would be done")
@click.option("--resolution", type=str, help="the resolution mode to pass to `uv sync`")
@click.argument("posargs", type=str, required=False, nargs=-1)
@click.pass_context
def cmd_uv_run(  # noqa: PLR0913  # this function does many things
    ctx: click.Context,
    *,
    env: str,
    force: bool,
    noop: bool,
    resolution: str | None,
    posargs: tuple[str],
) -> None:
    """Run tests using the uv tool."""
    try:
        cfg: Final = extract_cfg(ctx)
        proj: Final = parse.pyproject(cfg)
        posargs_list: Final = list(posargs)

        py_versions: Final = [cfg.python] if cfg.python != PY_SUPPORTED else proj.python_supported
        for pyver in py_versions:
            if pyver is None:
                cfg.log.info("Running the tests for the default Python version")
            else:
                cfg.log.info("Running the tests for Python %(pyver)s", {"pyver": pyver})

            runner = uv.UvRunner(
                cfg=dataclasses.replace(cfg, python=pyver),
                proj=proj,
                posargs=posargs_list,
                resolution=resolution,
            )
            runner.generate(check=False, diff=False, force=force, noop=noop)
            runner.run(envs=env.split(",") if env is not None else proj.envlist, noop=noop)
    except defs.Error as err:
        sys.exit(f"Could not run the tests using uv: {err}")


@click.group(name="uv")
@click.pass_context
def cmd_uv(ctx: click.Context) -> None:
    """Run tests using the uv tool."""


cmd_uv.add_command(cmd_uv_run)


@click.group(name="uvoxen")
@click.option(
    "--features",
    is_flag=True,
    is_eager=True,
    callback=arg_features,
    help="display program features information and exit",
)
@click.option(
    "--python",
    "-p",
    type=str,
    help=f"the Python version to use, or '{PY_SUPPORTED}' for all",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="quiet operation; only display warning and error messages",
)
@click.option("--verbose", "-v", is_flag=True, help="verbose operation; display diagnostic output")
@click.pass_context
def main(
    ctx: click.Context,
    *,
    features: bool,
    python: str | None,
    quiet: bool,
    verbose: bool,
) -> None:
    """Generate test configuration files and run tests."""
    if features:
        sys.exit("Internal error: how did we get to main() with features=True?")

    ctx.ensure_object(ConfigHolder)
    ctx.obj.cfg = defs.Config(
        log=util.build_logger(quiet=quiet, verbose=verbose),
        python=python,
        verbose=verbose,
    )


main.add_command(cmd_list)
main.add_command(cmd_req)
main.add_command(cmd_tox)
main.add_command(cmd_uv)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Test the Tox expansion emulator."""

from __future__ import annotations

import dataclasses
import pathlib
import shlex
import typing

from uvoxen import expand_tox_vars as etv


if typing.TYPE_CHECKING:
    from typing import Final


ENV_TRIVIAL: Final = {
    "USERNAME": "jrl",
}
"""Some simulated environment variables."""

GLOBALS_TRIVIAL: Final[dict[str, str | list[str]]] = {
    "toxinidir": "/nonexistent",
    "envdir": "/nosuch/env",
    "posargs": [],
}
"""Some simulated Tox global variables."""

GLOBALS_TRIVIAL_EXPANDED: Final[dict[str, str]] = {
    f"{{{name}}}": value if isinstance(value, str) else shlex.join(value)
    for name, value in GLOBALS_TRIVIAL.items()
}
"""The cached version of the Tox global variables."""

TRIVIAL_VARS_EXPANDED: Final = {
    "env": ENV_TRIVIAL,
    "setenv": {},
    "tox_globals": GLOBALS_TRIVIAL,
    "user_vars": {},
    "cache": GLOBALS_TRIVIAL_EXPANDED,
}


def test_empty() -> None:
    """No variables or anything."""
    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={},
        tox_globals=dict(GLOBALS_TRIVIAL),
        user_vars={},
    )

    # Make sure the `__post_init__()` method did not do anything rash."""
    assert dataclasses.asdict(tvars) == TRIVIAL_VARS_EXPANDED


def test_already_cached() -> None:
    """Look things up in the pre-generated cache."""
    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={},
        tox_globals=dict(GLOBALS_TRIVIAL),
        user_vars={},
    )
    assert dataclasses.asdict(tvars) == TRIVIAL_VARS_EXPANDED

    assert tvars.expand("{toxinidir}") == GLOBALS_TRIVIAL["toxinidir"]
    assert dataclasses.asdict(tvars) == TRIVIAL_VARS_EXPANDED


def run_expansion_list(tvars: etv.Variables, steps: list[tuple[str, str, dict[str, str]]]) -> None:
    """Expand the strings one by one, make sure the cache is updated."""
    cache = dict(tvars.cache)
    for line, expanded, added in steps:
        assert tvars.expand(line) == expanded, repr(tvars)
        assert tvars.cache == {**cache, **added}
        cache = dict(tvars.cache)


def test_non_recursive() -> None:
    """Expand a couple of single variables without recursion."""
    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={},
        tox_globals=dict(GLOBALS_TRIVIAL),
        user_vars={
            "defs": {"pyfiles": ["src/projname", "tests/unit"]},
            "something": {"else": "nothing-really"},
        },
    )
    assert tvars.cache == GLOBALS_TRIVIAL_EXPANDED

    no_vars: Final = "no variables at all"
    something_else: Final = "{[something]else}"
    nothing_really: Final = "nothing-really"
    sentence: Final = f"This is {something_else}, is it not?"
    sentence_exp: Final = f"This is {nothing_really}, is it not?"
    defs_pyfiles: Final = "{[defs]pyfiles}"
    src_files: Final = "src/projname tests/unit"
    cmd: Final = f"ruff --check {defs_pyfiles}"
    cmd_exp: Final = f"ruff --check {src_files}"
    run_expansion_list(
        tvars,
        [
            (no_vars, no_vars, {no_vars: no_vars}),
            (no_vars, no_vars, {}),
            (
                sentence,
                sentence_exp,
                {
                    sentence: sentence_exp,
                    nothing_really: nothing_really,
                    something_else: nothing_really,
                },
            ),
            (no_vars, no_vars, {}),
            (something_else, nothing_really, {}),
            (no_vars, no_vars, {}),
            (
                cmd,
                cmd_exp,
                {
                    cmd: cmd_exp,
                    defs_pyfiles: src_files,
                    "src/projname": "src/projname",
                    "tests/unit": "tests/unit",
                },
            ),
            (defs_pyfiles, src_files, {}),
            (sentence, sentence_exp, {}),
        ],
    )


def test_expand_recursively() -> None:  # noqa: PLR0914  # this is weird
    """Expand some variables recursively."""
    cmd_ruff = "{[commands]ruff}"
    cmd_ruff_exp: Final = "ruff --config ruff-base.toml"
    cmd_ruff_format: Final = "{[commands]ruff_format}"
    cmd_ruff_format_exp: Final = f"{cmd_ruff_exp} format"

    contrib_something_else: Final = "contrib/something-else"
    src_projname: Final = "src/projname"
    src_selftest: Final = "src/selftest"
    tests_unit: Final = "tests/unit"

    defs_pyfiles_mypy: Final = "{[defs]pyfiles_mypy}"
    defs_pyfiles_mypy_exp: Final = f"{src_projname} {src_selftest}"
    defs_pyfiles_ruff: Final = "{[defs]pyfiles_ruff}"
    defs_pyfiles_ruff_exp: Final = tests_unit
    defs_pyfiles_black: Final = "{[defs]pyfiles_black}"
    defs_pyfiles_black_exp: Final = f"{defs_pyfiles_mypy_exp} {defs_pyfiles_ruff_exp}"
    defs_pyfiles: Final = "{[defs]pyfiles}"
    defs_pyfiles_exp: Final = f"{defs_pyfiles_black_exp} {contrib_something_else}"

    cmd_ruff_format_pyfiles: Final = f"{cmd_ruff_format} {defs_pyfiles}"
    cmd_ruff_format_pyfiles_exp: Final = (
        f"{cmd_ruff_exp} format {src_projname} {src_selftest} {tests_unit} {contrib_something_else}"
    )

    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={},
        tox_globals=dict(GLOBALS_TRIVIAL),
        user_vars={
            "defs": {
                "pyfiles": [defs_pyfiles_black, contrib_something_else],
                "pyfiles_black": [defs_pyfiles_mypy, defs_pyfiles_ruff],
                "pyfiles_mypy": [src_projname, src_selftest],
                "pyfiles_ruff": [tests_unit],
            },
            "commands": {
                "ruff": cmd_ruff_exp,
                "ruff_format": f"{cmd_ruff} format",
            },
        },
    )
    assert tvars.cache == GLOBALS_TRIVIAL_EXPANDED

    run_expansion_list(
        tvars,
        [
            (
                cmd_ruff_format,
                cmd_ruff_format_exp,
                {
                    cmd_ruff: cmd_ruff_exp,
                    cmd_ruff_exp: cmd_ruff_exp,
                    cmd_ruff_format: cmd_ruff_format_exp,
                    f"{cmd_ruff} format": cmd_ruff_format_exp,
                },
            ),
            (
                cmd_ruff_format_pyfiles,
                cmd_ruff_format_pyfiles_exp,
                {
                    cmd_ruff_format_pyfiles: cmd_ruff_format_pyfiles_exp,
                    contrib_something_else: contrib_something_else,
                    defs_pyfiles: defs_pyfiles_exp,
                    defs_pyfiles_black: defs_pyfiles_black_exp,
                    defs_pyfiles_mypy: defs_pyfiles_mypy_exp,
                    defs_pyfiles_ruff: defs_pyfiles_ruff_exp,
                    src_projname: src_projname,
                    src_selftest: src_selftest,
                    tests_unit: tests_unit,
                },
            ),
        ],
    )


def test_expand_tox_too() -> None:
    """Also expand Tox variables."""
    defs_pyfiles: Final = "{[defs]pyfiles}"
    defs_pyfiles_exp: Final = "tests/unit"
    cmd_pytest: Final = "{[commands]pytest}"
    posargs: Final = "{posargs}"
    posargs_exp: Final = "-vv -k 'test_expand and more'"
    cmd_pytest_cmd: Final = f"pytest {posargs} {defs_pyfiles}"
    cmd_pytest_exp: Final = f"pytest {posargs_exp} {defs_pyfiles_exp}"

    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={},
        tox_globals=etv.prepare_tox_vars(
            toxinidir=pathlib.Path("/nonexistent"),
            envdir=pathlib.Path("/nonesuch/venv"),
            posargs=["-vv", "-k", "test_expand and more"],
        ),
        user_vars={
            "commands": {
                "pytest": cmd_pytest_cmd,
            },
            "defs": {
                "pyfiles": defs_pyfiles_exp,
            },
        },
    )
    assert tvars.cache[posargs] == posargs_exp

    run_expansion_list(
        tvars,
        [
            (
                cmd_pytest,
                cmd_pytest_exp,
                {
                    cmd_pytest: cmd_pytest_exp,
                    cmd_pytest_cmd: cmd_pytest_exp,
                    defs_pyfiles: defs_pyfiles_exp,
                    defs_pyfiles_exp: defs_pyfiles_exp,
                },
            ),
            (defs_pyfiles, defs_pyfiles_exp, {}),
        ],
    )


def test_setenv() -> None:
    """Test `setenv` processing."""
    mypypath: Final = "{toxinidir}/stubs and more"
    mypypath_exp: Final = f"{GLOBALS_TRIVIAL['toxinidir']}/stubs and more"
    typeshed: Final = "{env:MYPYPATH}/typeshed"
    typeshed_exp: Final = f"{mypypath_exp}/typeshed"

    tvars: Final = etv.Variables(
        env=dict(ENV_TRIVIAL),
        setenv={
            "MYPYPATH": mypypath,
            "TYPESHED": typeshed,
        },
        tox_globals=dict(GLOBALS_TRIVIAL),
        user_vars={},
    )
    assert tvars.cache == {
        **GLOBALS_TRIVIAL_EXPANDED,
        "{env:MYPYPATH}": mypypath_exp,
        "{env:MYPYPATH}/typeshed": typeshed_exp,
        "{env:TYPESHED}": typeshed_exp,
        mypypath: mypypath_exp,
    }

    run_expansion_list(
        tvars,
        [
            ("{env:MYPYPATH}", mypypath_exp, {}),
            ("{env:TYPESHED}", typeshed_exp, {}),
            (
                "{env:NONESUCH}",
                "",
                {
                    "{env:NONESUCH}": "",
                },
            ),
        ],
    )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Run commands for the specified environments."""

from __future__ import annotations

import dataclasses
import pathlib
import shlex
import subprocess  # noqa: S404
import typing

from uvoxen import defs
from uvoxen import expand_tox_vars
from uvoxen import parse


if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Final


@dataclasses.dataclass
class NotRegularFileError(defs.Error):
    """The specified output path points to something that is not a regular file."""

    path: pathlib.Path
    """The path to what we expected to be a file."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Not a regular file: {self.path}"


@dataclasses.dataclass
class NeedsRegeneratingError(defs.Error):
    """One or more test configuration files need to be regenerated."""

    paths: list[pathlib.Path]
    """The paths to the files that need to be regenerated."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Need to regenerate {' '.join(str(path) for path in self.paths)}"


@dataclasses.dataclass
class CommandError(defs.Error):
    """An error that occurred when running a command."""

    cmd: list[str]
    """The command we tried to run."""

    @property
    def cmdstr(self) -> str:
        """A string representation of the command we tried to run."""
        return shlex.join(self.cmd)

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Error running the `{self.cmdstr}` command: {self!r}"


@dataclasses.dataclass
class CommandRunError(CommandError):
    """Could not run the specified command at all."""

    err: Exception
    """The error that occurred."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Could not run the `{self.cmdstr}` command: {self.err}"


@dataclasses.dataclass
class CommandFailedError(CommandError):
    """The specified command exited with a non-zero code."""

    code: int
    """The exit code."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"The `{self.cmdstr}` command exited with code {self.code}"


def prepare_tox_vars(*, posargs: list[str]) -> dict[str, str | list[str]]:
    """Simulate (to some extent) the way Tox prepares its environment variables."""
    toxinidir: Final = pathlib.Path.cwd()
    return expand_tox_vars.prepare_tox_vars(
        toxinidir=toxinidir,
        envdir=toxinidir / ".venv",
        posargs=posargs,
    )


@dataclasses.dataclass(frozen=True)
class Runner:
    """Abstract base class for running commands for environments."""

    cfg: defs.Config
    """The runtime configuration."""

    posargs: list[str]
    """The command-line arguments passed for this run."""

    proj: parse.Project
    """The project to run environments for."""

    def _generate_file_check_only(
        self,
        *,
        output: str,
        opath: pathlib.Path,
        contents: str,
        diff: bool,
        is_up_to_date: Callable[[], bool],
    ) -> None:
        """Handle check-only mode according to the other options."""
        if is_up_to_date():
            self.cfg.log.info("The %(output)s file is up to date", {"output": output})
            return

        if diff:
            subprocess.run(  # noqa: S603
                ["diff", "-u", "--", opath, "-"],  # noqa: S607
                check=False,
                encoding="UTF-8",
                input=contents,
            )

        raise NeedsRegeneratingError([opath])

    def generate_file(  # noqa: PLR0913  # this function does many things
        self,
        *,
        check: bool,
        diff: bool,
        force: bool,
        noop: bool,
        output: str,
        contents: str,
    ) -> None:
        """Write a file out if needed."""
        if output == "-":
            print(contents, end="")  # noqa: T201
            return

        opath: Final = pathlib.Path(output)
        if opath.is_symlink() or (opath.exists() and not opath.is_file()):
            raise NotRegularFileError(opath)

        def is_up_to_date() -> bool:
            """Check whether the file contents are up to date."""
            if not opath.exists():
                return False

            try:
                return opath.read_text(encoding="UTF-8") == contents
            except ValueError:
                return False

        if check:
            if not force:
                self._generate_file_check_only(
                    contents=contents,
                    output=output,
                    opath=opath,
                    diff=diff,
                    is_up_to_date=is_up_to_date,
                )
                return
        elif is_up_to_date():
            self.cfg.log.info("The %(output)s file is up to date", {"output": output})
            return

        self.cfg.log.info(
            "Writing %(count)d characters to the %(output)s file",
            {"count": len(contents), "output": output},
        )
        if noop:
            self.cfg.log.info("Would write to %(output)s", {"output": output})
        else:
            opath.write_text(contents, encoding="UTF-8")

    def generate(
        self,
        *,
        check: bool,
        diff: bool,
        force: bool,
        noop: bool = False,
        output: str | None = None,
    ) -> None:
        """Generate the test configuration if necessary."""
        raise NotImplementedError(repr(self))

    def run(self, *, envs: list[str], noop: bool = False) -> None:
        """Generate and run commands."""
        raise NotImplementedError(repr(self))

    def run_command(
        self,
        cmd: list[str],
        *,
        env: dict[str, str] | None = None,
        noop: bool = False,
    ) -> None:
        """Run a single command or display it in noop mode."""
        cmdstr: Final = shlex.join(cmd)
        if noop:
            self.cfg.log.info("Would run `%(cmdstr)s`", {"cmdstr": cmdstr})
            print(cmdstr)  # noqa: T201
            return

        self.cfg.log.info("About to run `%(cmdstr)s`", {"cmdstr": cmdstr})
        if env is not None:
            self.cfg.log.info("- environment: %(env)r", {"env": env})
        try:
            subprocess.check_call(cmd, env=env, shell=False)  # noqa: S603
        except OSError as err:
            raise CommandRunError(cmd, err) from err
        except subprocess.CalledProcessError as err:
            raise CommandFailedError(cmd, err.returncode) from err



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Simulate some of Tox's variable expansion."""

from __future__ import annotations

import dataclasses
import os
import os.path
import shlex
import typing

import pyparsing as pyp


if typing.TYPE_CHECKING:
    import pathlib
    from typing import Final

    from . import parse


ENV_PRESERVE_NAMES = (
    "https_proxy",
    "http_proxy",
    "no_proxy",
    "LANG",
    "LANGUAGE",
    "CURL_CA_BUNDLE",
    "SSL_CERT_FILE",
    "CC",
    "CFLAGS",
    "CCSHARED",
    "CXX",
    "CPPFLAGS",
    "LD_LIBRARY_PATH",
    "LDFLAGS",
    "HOME",
    "FORCE_COLOR",
    "NO_COLOR",
    "TMPDIR",
    "TEMP",
    "TMP",
    "USERPROFILE",
    "PATHEXT",
    "MSYSTEM",
    "WINDIR",
    "APPDATA",
    "PROGRAMDATA",
    "PROGRAMFILES(x86)",
    "SYSTEMDRIVE",
    "SYSTEMROOT",
    "COMSPEC",
    "PROCESSOR_ARCHITECTURE",
    "NUMBER_OF_PROCESSORS",
    "NETRC",
    "NIX_LD_LIBRARY_PATH",
    "TERM",
    "LINES",
    "COLUMNS",
)

ENV_PRESERVE_PREFIXES = (
    "PIP_",
    "VIRTUALENV_",
    "NIX_LD",
    "UV_",
)


@dataclasses.dataclass(frozen=True)
class ParsedPart:
    """The base class for parsed parts."""

    raw: str
    """The raw part."""


@dataclasses.dataclass(frozen=True)
class ParsedLiteral(ParsedPart):
    """A literal string."""

    value: str
    """The literal string itself."""


@dataclasses.dataclass(frozen=True)
class ParsedIdent(ParsedPart):
    """An identifier."""

    name: str
    """The identifier name."""


@dataclasses.dataclass(frozen=True)
class ParsedUserVar(ParsedPart):
    """A user variable with a section name."""

    section: str
    """The section name."""

    name: str
    """The variable name within the section."""


@dataclasses.dataclass(frozen=True)
class ParsedToxVar(ParsedPart):
    """A Tox variable without a section name."""

    name: str
    """The variable name."""


@dataclasses.dataclass(frozen=True)
class ParsedEnvVar(ParsedPart):
    """An environment variable."""

    name: str
    """The variable name."""


def raw_parts_to_str(tokens: pyp.ParseResults) -> str:
    """Build the raw string out of the parsed tokens."""

    def single(part: ParsedPart | str) -> str:
        """Get the raw string from a single parsed token."""
        match part:
            case ParsedPart(raw):
                return raw

            case str(raw_str):
                return raw_str

            case other:
                raise RuntimeError(repr(other))

    return "".join(single(part) for part in tokens.as_list())


_p_ident = pyp.Word(pyp.srange("[A-Za-z0-9_]"))


@_p_ident.set_parse_action
def _process_ident(tokens: pyp.ParseResults) -> ParsedIdent:
    """Build an identifier out of its name."""
    raw: Final = raw_parts_to_str(tokens)
    res: Final = tokens.as_list()
    match res:
        case [ident]:
            return ParsedIdent(raw, ident)

        case other:
            raise RuntimeError(repr(other))


_p_literal = pyp.CharsNotIn("{")


@_p_literal.set_parse_action
def _process_literal(tokens: pyp.ParseResults) -> ParsedLiteral:
    """Record the literal string."""
    raw: Final = raw_parts_to_str(tokens)
    res: Final = tokens.as_list()
    match res:
        case [str(literal)]:
            return ParsedLiteral(raw, literal)

        case other:
            raise RuntimeError(repr((tokens, other)))


_p_user_var = pyp.Literal("{[") + _p_ident + pyp.Literal("]") + _p_ident + pyp.Literal("}")


@_p_user_var.set_parse_action
def _process_user_var(tokens: pyp.ParseResults) -> ParsedUserVar:
    """Build a user var out of the identifiers."""
    raw: Final = raw_parts_to_str(tokens)
    res: Final = tokens.as_list()
    match res:
        case [_, ParsedIdent(_, section), _, ParsedIdent(_, name), _]:
            return ParsedUserVar(raw, section, name)

        case other:
            raise RuntimeError(repr(other))


_p_tox_var = pyp.Literal("{") + _p_ident + pyp.Literal("}")


@_p_tox_var.set_parse_action
def _process_tox_var(tokens: pyp.ParseResults) -> ParsedToxVar:
    """Build a Tox var out of the identifier."""
    raw: Final = raw_parts_to_str(tokens)
    res: Final = tokens.as_list()
    match res:
        case [_, ParsedIdent(_, name), _]:
            return ParsedToxVar(raw, name)

        case other:
            raise RuntimeError(repr(other))


_p_env_var = pyp.Literal("{env:") + _p_ident + pyp.Literal("}")


@_p_env_var.set_parse_action
def _process_env_var(tokens: pyp.ParseResults) -> ParsedEnvVar:
    """Build a Tox var out of the identifier."""
    raw: Final = raw_parts_to_str(tokens)
    res: Final = tokens.as_list()
    match res:
        case [_, ParsedIdent(_, name), _]:
            return ParsedEnvVar(raw, name)

        case other:
            raise RuntimeError(repr(other))


_p_part = _p_user_var | _p_tox_var | _p_env_var | _p_literal

_p_full_line = pyp.ZeroOrMore(_p_part)

_p_full_line_complete: Final = _p_full_line.leave_whitespace()


def quote_list(values: list[str]) -> str:
    """Quote the values using `shlex`.

    Break this out into a separate function in case we change our mind in
    the future of whether we want the quoting to happen here or not.
    """
    return " ".join(values)


@dataclasses.dataclass
class Variables:
    """The list of definitions, environment variables, and cached expansions."""

    env: dict[str, str]
    """The environment variables that this environment will see (before `setenv` processing)."""

    setenv: dict[str, str]
    """The `setenv` directives for this environment."""

    tox_globals: dict[str, str | list[str]]
    """The global variables, e.g. `posargs`, `envdir`, etc."""

    user_vars: dict[str, dict[str, str | list[str]]]
    """Variables defined in the `tox.ini` sections."""

    cache: dict[str, str] = dataclasses.field(default_factory=dict)
    """Cached values for variables that have already been expanded."""

    def __post_init__(self) -> None:
        """Expand the `setenv` definitions, cache `tox_globals`."""
        for name, value in self.tox_globals.items():
            exp_name = f"{{{name}}}"
            if exp_name in self.cache:
                raise RuntimeError(repr((self, name, value, exp_name)))

            match value:
                case str(value_str):
                    self.cache[exp_name] = value_str

                case list(value_list):
                    self.cache[exp_name] = quote_list(value_list)

                case value_other:
                    raise RuntimeError(repr((self, name, value, value_other)))

        for name, value in self.setenv.items():
            exp_name = f"{{env:{name}}}"
            if exp_name in self.cache:
                # Already referenced by another `setenv` variable
                continue

            self.env[name] = self.expand(exp_name)
            if exp_name not in self.cache:
                raise RuntimeError(repr((self, name, value, exp_name)))

    def _expand_single(self, part: ParsedPart) -> str:  # noqa: C901  # it is what it is
        """Expand a single part of the input line."""

        def cache_and_return(value: str) -> str:
            """Store the value into the cache and return it."""
            self.cache[part.raw] = value
            return value

        def scalar_or_list(value: str | list[str]) -> str:
            """Expand a value."""
            match value:
                case str(value_str):
                    return cache_and_return(self.expand(value_str))

                case list(value_list):
                    return cache_and_return(
                        " ".join(self.expand(item) for item in value_list),
                    )

                case value_other:
                    raise RuntimeError(repr((self, part, value_other)))

        if part.raw in self.cache:
            return self.cache[part.raw]

        match part:
            case ParsedLiteral(_literal_raw, literal_value):
                return literal_value

            case ParsedUserVar(_var_raw, var_section, var_name):
                return scalar_or_list(self.user_vars[var_section][var_name])

            case ParsedToxVar(_var_raw, var_name):
                return scalar_or_list(self.tox_globals[var_name])

            case ParsedEnvVar(_var_raw, var_name):
                setenv_value: Final = self.setenv.get(var_name)
                if setenv_value is not None:
                    return scalar_or_list(setenv_value)

                env_value: Final = self.env.get(var_name, "")
                return cache_and_return(env_value)

            case other:
                raise RuntimeError(repr((self, other)))

    def expand(self, line: str) -> str:
        """Expand (perhaps recursively) variables within a line."""
        if line in self.cache:
            return self.cache[line]

        parsed_parts: Final = _p_full_line_complete.parse_string(line, parse_all=True).as_list()
        res: Final = "".join(self._expand_single(part) for part in parsed_parts)
        self.cache[line] = res
        return res


def prepare_env(*, env: parse.Environment, envdir: pathlib.Path) -> dict[str, str]:
    """Sanitize the current environment."""
    return {
        name: value
        for name, value in os.environ.items()
        if name in ENV_PRESERVE_NAMES
        or name.startswith(ENV_PRESERVE_PREFIXES)
        or name in env.passenv
    } | {"PATH": f"{envdir}/bin{os.path.pathsep}{os.environ['PATH']}"}


def prepare_tox_vars(
    *,
    toxinidir: pathlib.Path,
    envdir: pathlib.Path,
    posargs: list[str],
) -> dict[str, str | list[str]]:
    """Simulate (to some extent) the way Tox prepares its environment variables."""
    return {
        "envdir": str(envdir),
        "posargs": [shlex.quote(word) for word in posargs],
        "toxinidir": str(toxinidir),
    }



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Configure and run tests using the `tox` tool."""

from __future__ import annotations

import dataclasses
import itertools
import typing

from . import base


if typing.TYPE_CHECKING:
    from typing import Final

    from uvoxen import parse


@dataclasses.dataclass(frozen=True)
class ToxRunner(base.Runner):
    """Generate `tox.ini`, run the tests using Tox."""

    def expand_tox_defs(self) -> str:
        """Expand the `[defs]` section: strings or lists of strings."""

        def single(name: str, value: str | list[str]) -> str:
            """Expand the value of a single definition."""
            match value:
                case str(value_str):
                    return f"{name} = {value_str}"

                case list(values):
                    lines: Final = " \\\n".join(f"  {item}" for item in values)
                    return f"{name} =\n{lines}"

                case other:
                    raise RuntimeError(repr(other))

        if not self.proj.defs:
            return ""

        expanded: Final = "\n".join(itertools.starmap(single, self.proj.defs.items()))
        return f"[defs]\n{expanded}"

    def get_tox_env_order(self) -> list[str]:
        """Get the order of the environments to write into the Tox configuration."""

        def listed_before_others(listed: list[str], envs: list[str]) -> list[str]:
            """Put the listed environments first."""
            return [*listed, *(name for name in envs if name not in listed)]

        return listed_before_others(
            self.proj.tox.env_order
            if self.proj.tox is not None and self.proj.tox.env_order is not None
            else self.proj.envlist,
            sorted(self.proj.env),
        )

    def generate_env(self, name: str, env: parse.Environment) -> str:
        """Generate the configuration for a single Tox environment."""
        commands: Final = "\n".join(f"  {cmd}" for cmd in env.commands)

        skip_install = (
            ""
            if not self.proj.build_project or env.install_project or env.install_dependencies
            else "\nskip_install = True"
        )

        tags = "" if env.tags is None else "\ntags =\n" + "\n".join(f"  {tag}" for tag in env.tags)

        dep_groups = (
            "\ndependency_groups =\n" + "\n".join(f"  {dep}" for dep in env.dependency_groups)
            if env.dependency_groups
            else ""
        )

        setenv = (
            "\nsetenv =\n"
            + "\n".join(f"  {name} = {value}" for name, value in sorted(env.setenv.items()))
            if env.setenv
            else ""
        )

        passenv = (
            "\npassenv =\n" + "\n".join(f"  {name}" for name in sorted(env.passenv))
            if env.passenv
            else ""
        )

        externals = (
            "\nallowlist_externals =\n" + "\n".join(f"  {item}" for item in env.allowlist_externals)
            if env.allowlist_externals is not None
            else ""
        )

        return f"""
[testenv:{name}]{skip_install}{tags}{dep_groups}{setenv}{passenv}{externals}
commands =
{commands}
"""

    def generate(
        self,
        *,
        check: bool,
        diff: bool,
        force: bool,
        noop: bool = False,
        output: str | None = None,
    ) -> None:
        """Write `tox.ini` out if necessary."""
        header: Final = "" if self.proj.tox is None else self.proj.tox.header
        footer: Final = "" if self.proj.tox is None else self.proj.tox.footer
        isolated_or_sdist = (
            "isolated_build = True" if self.proj.build_project else "skipsdist = True"
        )
        envlist: Final = "".join(
            f"\n  {name}"
            for name in [
                *self.proj.envlist,
                *(
                    self.proj.tox.envlist_append
                    if self.proj.tox is not None and self.proj.tox.envlist_append is not None
                    else []
                ),
            ]
        )
        defs_expanded: Final = self.expand_tox_defs()
        envs: Final = "".join(
            itertools.starmap(
                self.generate_env,
                ((name, self.proj.env[name]) for name in self.get_tox_env_order()),
            ),
        )
        res = f"""{header}[tox]
minversion = 4.22
envlist ={envlist}
{isolated_or_sdist}

{defs_expanded}
{envs}{footer}
"""[:-1]

        self.generate_file(
            check=check,
            diff=diff,
            force=force,
            noop=noop,
            output="tox.ini" if output is None else output,
            contents=res,
        )

    def run(self, *, envs: list[str], noop: bool = False) -> None:
        """Run `tox`."""
        self.run_command(
            [
                "uvx",
                *(["-p", self.cfg.python] if self.cfg.python is not None else []),
                "--with",
                "tox-uv",
                "--",
                "tox",
                "run-parallel",
                *(["-e", ",".join(envs)] if envs else []),
                "--",
                *self.posargs,
            ],
            noop=noop,
        )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Basic test for file importing."""

from __future__ import annotations

import pathlib
import re
import sys
import typing

import feature_check
from packaging import version as pkg_version

from uvoxen import defs


_RE_FEATURE_VALUE: Final = re.compile(
    r"^ (?P<major> 0 | [1-9][0-9]* ) \. (?P<minor> 0 | [1-9][0-9]* )",
    re.X,
)


if typing.TYPE_CHECKING:
    from typing import Final


def test_version() -> None:
    """Make sure the `VERSION` variable has a sane value."""
    version: Final = pkg_version.Version(defs.VERSION)
    assert version > pkg_version.Version("0")


def test_features() -> None:
    """Make sure that the list of features looks right.

    It must include the program's name, and each value must be a X.Y number pair.
    """
    assert defs.FEATURES["uvoxen"] == defs.VERSION
    for value in (value for name, value in defs.FEATURES.items() if name != "uvoxen"):
        assert _RE_FEATURE_VALUE.match(value)


def test_features_run() -> None:
    """Run the command-line tool with the `--features` option, see what happens."""
    test_program: Final = pathlib.Path(sys.executable).parent / "uvoxen"
    features: Final = feature_check.obtain_features(str(test_program))
    assert features == {
        name: feature_check.parse_version(version) for name, version in defs.FEATURES.items()
    }



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Configure and run tests using the `uv` tool."""

from __future__ import annotations

import dataclasses
import itertools
import pathlib
import shlex
import typing

from uvoxen import defs
from uvoxen import expand_tox_vars

from . import base


if typing.TYPE_CHECKING:
    from typing import Final


@dataclasses.dataclass
class EnvUndefinedError(defs.Error):
    """One or more of the specified environments are not defined."""

    missing_envs: list[str]
    """The missing environments."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        plu: Final = "" if len(self.missing_envs) == 1 else "s"
        return f"Undefined environment{plu}: {' '.join(self.missing_envs)}"


@dataclasses.dataclass(frozen=True)
class UvRunner(base.Runner):
    """Run commands using the `uv` tool."""

    resolution: str | None
    """The resolution mode to pass to `uv sync`."""

    def generate(
        self,
        *,
        check: bool,  # noqa: ARG002  # class inheritance
        diff: bool,  # noqa: ARG002  # class inheritance
        force: bool,  # noqa: ARG002  # class inheritance
        noop: bool = False,  # noqa: ARG002  # class inheritance
        output: str | None = None,  # noqa: ARG002  # class inheritance
    ) -> None:
        """Generate the test configuration if necessary."""
        self.cfg.log.info("Nothing to generate for `uv`")

    def run(self, *, envs: list[str], noop: bool = False) -> None:
        """Run commands."""
        missing_envs: Final = [env for env in envs if env not in self.proj.env]
        if missing_envs:
            raise EnvUndefinedError(missing_envs)

        for name, env in ((name, self.proj.env[name]) for name in envs):
            tox_globals = base.prepare_tox_vars(posargs=self.posargs)
            tox_env = expand_tox_vars.prepare_env(
                env=env,
                envdir=pathlib.Path(str(tox_globals["envdir"])),
            )
            self.cfg.log.debug("- prepared tox vars %(tox_globals)r", {"tox_globals": tox_globals})
            tvars = expand_tox_vars.Variables(
                env=tox_env,
                setenv=env.setenv,
                tox_globals=tox_globals,
                user_vars={"defs": self.proj.defs},
            )

            self.cfg.log.info("Running commands for the %(name)s environment", {"name": name})

            self.run_command(["rm", "-rf", "--", ".venv"], noop=noop)

            # sync
            self.run_command(
                [
                    "uv",
                    "sync",
                    *([] if self.cfg.python is None else ["-p", self.cfg.python]),
                    *(
                        []
                        if self.proj.build_project and env.install_project
                        else ["--no-install-project"]
                    ),
                    *([] if self.resolution is None else ["--resolution", self.resolution]),
                    *itertools.chain(
                        *(["--group", dep_group] for dep_group in env.dependency_groups),
                    ),
                ],
                noop=noop,
            )

            # run
            for cmd_raw in env.commands:
                cmd_exp = tvars.expand(cmd_raw)
                self.run_command(
                    ["uv", "run", "--no-sync", "--", *shlex.split(cmd_exp)],
                    env=tvars.env,
                    noop=noop,
                )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Generate `requirements/*.txt` files out of the dependency group definitions."""

from __future__ import annotations

import dataclasses
import typing

from uvoxen import parse

from . import base


if typing.TYPE_CHECKING:
    from typing import Final


@dataclasses.dataclass(frozen=True)
class ReqRunner(base.Runner):
    """Generate `requirements/*.txt` files out of the dependency group definitions."""

    group: str
    """The dependency grouop to generate the requirements file for."""

    def generate(
        self,
        *,
        check: bool,
        diff: bool,
        force: bool,
        noop: bool = False,
        output: str | None = None,
    ) -> None:
        """Write the requirements file out if necessary."""

        def single(item: parse.PyDependencyGroupItem) -> str:
            """Output one or more lines corresponding to this dependency group item."""
            match item:
                case parse.PyDependencyGroupItemReq(req):
                    return f"{req}\n"

                case parse.PyDependencyGroupIncluded(group):
                    return all_items(group)

                case other:
                    raise RuntimeError(repr(other))

        def all_items(group: str) -> str:
            """Output one or more lines corresponding to these dependency group items."""
            return "".join(single(item) for item in self.proj.python_dependency_groups[group])

        header: Final = (
            "" if self.proj.req is None or self.proj.req.header is None else self.proj.req.header
        )

        footer: Final = (
            "" if self.proj.req is None or self.proj.req.footer is None else self.proj.req.footer
        )

        items: Final = all_items(self.group)

        res = f"{header}{items}{footer}"

        return self.generate_file(
            check=check,
            diff=diff,
            force=force,
            noop=noop,
            output=f"requirements/{self.group}.txt" if output is None else output,
            contents=res,
        )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Parse the pyproject.toml configuration."""

from __future__ import annotations

import dataclasses
import functools
import pathlib
import re
import tomllib
import typing

import mashumaro
from mashumaro import codecs as m_codecs
from mashumaro import config as m_config

from . import defs


if typing.TYPE_CHECKING:
    from typing import Any, Final


@dataclasses.dataclass
class ParseError(defs.Error):
    """An error that occurred while parsing the `pyproject.toml` configuration."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Error parsing `pyproject.toml`: {self!r}"


@dataclasses.dataclass
class ProjMissingError(ParseError):
    """Missing values in the `pyproject.toml` file."""

    path: str
    """The missing value."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Missing {self.path} in pyproject.toml"


@dataclasses.dataclass
class ProjWrongTypeError(ParseError):
    """Wrong type for a `pyproject.toml` value."""

    path: str
    """The wrong value."""

    exp_type: str
    """The expected type for the value."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Expected the {self.path} pyproject.toml value to be a {self.exp_type}"


@dataclasses.dataclass
class ProjFormatVersionUnsupportedError(ParseError):
    """Unsupported format version in the `pyproject.toml` file."""

    major: int
    """The major version number from the file."""

    minor: int
    """The minor version number from the file."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Unsupported uvoxen format version {self.major}.{self.minor} in pyproject.toml"


@dataclasses.dataclass
class ProjDeserializeError(ParseError):
    """Could not deserialize the `uvoxen` configuration in the `pyproject.toml` file."""

    err: Exception
    """The error that occurred."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"Could not parse the tool.uvoxen pyproject.toml section: {self.err}"


RE_LINE_VARS: Final = re.compile(
    r"""
        \{
            \[ (?P<section> [A-Za-z0-9_]+ ) \]
            (?P<name> [A-Za-z0-9_]+ )
        \}
    """,
    re.X,
)

RE_PY_SUPPORTED: Final = re.compile(
    r"""
        ^
        Programming \s Language \s :: \s Python \s :: \s
        (?P<version>
          3\.
          (?: 0 | [1-9][0-9]* )
        )
        $
    """,
    re.X,
)


@dataclasses.dataclass(frozen=True)
class PyDependencyGroupItem:
    """The base class for a single item in a dependency group."""


@dataclasses.dataclass(frozen=True)
class PyDependencyGroupItemReq(PyDependencyGroupItem):
    """A requirement specification within a dependency group."""

    req: str
    """The requirement string."""


@dataclasses.dataclass(frozen=True)
class PyDependencyGroupIncluded(PyDependencyGroupItem):
    """The name of another dependency group to be included herewith."""

    group: str
    """The name of the other dependency group."""


@dataclasses.dataclass(frozen=True)
class Environment(mashumaro.DataClassDictMixin):
    """A single test environment."""

    commands: list[str]
    """The commands to run for this test environment."""

    dependency_groups: list[str] = dataclasses.field(
        default_factory=list,
        metadata=mashumaro.field_options(alias="dependency-groups"),
    )
    """The list of Python package groups to install in this environment."""

    install_dependencies: bool = dataclasses.field(
        default=False,
        metadata=mashumaro.field_options(alias="install-dependencies"),
    )
    """Whether to install the project dependencies."""

    install_project: bool = dataclasses.field(
        default=False,
        metadata=mashumaro.field_options(alias="install-project"),
    )
    """Whether to install the project itself along with its dependencies."""

    allowlist_externals: list[str] | None = dataclasses.field(
        default=None,
        metadata=mashumaro.field_options(alias="allowlist-externals"),
    )
    """The external commands to allow Tox to run."""

    setenv: dict[str, str] = dataclasses.field(default_factory=dict)
    """Additional environment variables to set."""

    passenv: list[str] = dataclasses.field(default_factory=list)
    """Environment variables to pass from the parent process environment."""

    tags: list[str] | None = None
    """The tags for ordering the environments into stages."""

    class Config(m_config.BaseConfig):
        """Make sure the parser fails on extraneous keys."""

        forbid_extra_keys = True


@dataclasses.dataclass(frozen=True)
class ToxSettings(mashumaro.DataClassDictMixin):
    """Tox-specific settings."""

    envlist_append: list[str] | None = dataclasses.field(
        default=None,
        metadata=mashumaro.field_options(alias="envlist-append"),
    )
    """Additional entries to add to `tox.envlist`."""

    env_order: list[str] | None = dataclasses.field(
        default=None,
        metadata=mashumaro.field_options(alias="env-order"),
    )
    """The order in which the environments should appear in the Tox configuration."""

    footer: str = ""
    """Text to place at the bottom of the file."""

    header: str = ""
    """Text to place at the top of the file."""

    class Config(m_config.BaseConfig):
        """Make sure the parser fails on extraneous keys."""

        forbid_extra_keys = True


@dataclasses.dataclass(frozen=True)
class ReqSettings(mashumaro.DataClassDictMixin):
    """Requirements-specific settings."""

    footer: str = ""
    """Text to place at the bottom of the file."""

    header: str = ""
    """Text to place at the top of the file."""

    class Config(m_config.BaseConfig):
        """Make sure the parser fails on extraneous keys."""

        forbid_extra_keys = True


@dataclasses.dataclass(frozen=True)
class ProjectParsed(mashumaro.DataClassDictMixin):
    """A set of test environments and definitions."""

    env: dict[str, Environment]
    """The actual test environments defined."""

    envlist: list[str]
    """The list of environments to run by default."""

    build_project: bool = dataclasses.field(
        default=True,
        metadata=mashumaro.field_options(alias="build-project"),
    )
    """Whether to build the Python project at all, or merely run tests."""

    defs: dict[str, str | list[str]] = dataclasses.field(default_factory=dict)
    """The variables to expand within the commands and the test configuration."""

    req: ReqSettings | None = None
    """Requirements-specific settings, if any."""

    tox: ToxSettings | None = None
    """Tox-specific settings, if any."""

    class Config(m_config.BaseConfig):
        """Make sure the parser fails on extraneous keys."""

        forbid_extra_keys = True


@dataclasses.dataclass(frozen=True)
class ProjectParsedV0dot1(mashumaro.DataClassDictMixin):
    """A set of test environments and definitions."""

    env: dict[str, Environment]
    """The actual test environments defined."""

    envlist: list[str]
    """The list of environments to run by default."""

    defs: dict[str, str | list[str]] = dataclasses.field(default_factory=dict)
    """The variables to expand within the commands and the test configuration."""

    req: ReqSettings | None = None
    """Requirements-specific settings, if any."""

    tox: ToxSettings | None = None
    """Tox-specific settings, if any."""

    class Config(m_config.BaseConfig):
        """Make sure the parser fails on extraneous keys."""

        forbid_extra_keys = True

    def upgrade_to_latest(self) -> ProjectParsed:
        """Upgrade the parsed configuration to the latest format."""
        return ProjectParsed(
            env=self.env,
            envlist=self.envlist,
            build_project=True,
            defs=self.defs,
            req=self.req,
            tox=self.tox,
        )


@dataclasses.dataclass(frozen=True)
class Project:
    """A set of test environments and definitions."""

    build_project: bool
    """Whether to build the Python project at all, or merely run tests."""

    defs: dict[str, str | list[str]]
    """The variables to expand within the commands and the test configuration."""

    env: dict[str, Environment]
    """The actual test environments defined."""

    envlist: list[str]
    """The list of environments to run by default."""

    python_dependency_groups: dict[str, list[PyDependencyGroupItem]]
    """The `dependency-groups` section of the `pyproject.toml` file."""

    python_supported: list[str]
    """The list of supported Python versions as declared in `pyproject.toml` trove classifiers."""

    req: ReqSettings | None = None
    """Requirements-specific settings, if any."""

    tox: ToxSettings | None = None
    """Tox-specific settings, if any."""


def _parse_dep_group(name: str, items: Any) -> list[PyDependencyGroupItem]:  # noqa: ANN401
    """Parse the list of items within a `pyproject.toml` dependency group."""
    if not isinstance(items, list):
        raise ProjWrongTypeError(f"dependency-groups.{name}", "table")  # noqa: EM102

    def single(item: Any) -> PyDependencyGroupItem:  # noqa: ANN401
        """Parse a single item within a dependency group."""
        match item:
            case str(item_req):
                return PyDependencyGroupItemReq(item_req)

            case dict(item_dict):
                included: Final = item_dict.get("include-group")
                match included:
                    case None:
                        raise ProjMissingError(
                            f"dependency-groups.{name}.{{include-group}}",  # noqa: EM102
                        )

                    case str(included_group):
                        return PyDependencyGroupIncluded(included_group)

                    case _:
                        raise ProjWrongTypeError(
                            f"dependency-group.{name}.{{item}}",  # noqa: EM102
                            "include-group",
                        )

            case _:
                raise ProjWrongTypeError(
                    f"dependency-group.{name}.item",  # noqa: EM102
                    "string or table",
                )

    return [single(item) for item in items]


@functools.lru_cache
def _pyproject_loader() -> m_codecs.BasicDecoder[ProjectParsed]:
    """Instantiate a typed loader that will parse the `tool.uvoxen` pyproject section."""
    return m_codecs.BasicDecoder(ProjectParsed)


@functools.lru_cache
def _pyproject_loader_v0dot1() -> m_codecs.BasicDecoder[ProjectParsedV0dot1]:
    """Instantiate a typed loader that will parse the `tool.uvoxen` pyproject section."""
    return m_codecs.BasicDecoder(ProjectParsedV0dot1)


def _pyproject_load(proj_tool: dict[str, Any]) -> ProjectParsed:
    """Parse the [tool.uvoxen] section according to its .format.version field."""
    try:
        ver_major: Final = proj_tool["format"]["version"]["major"]
        ver_minor: Final = proj_tool["format"]["version"]["minor"]
    except (KeyError, TypeError) as err:
        raise ProjMissingError("tool.uvoxen.version.{major,minor}") from err  # noqa: EM101
    if not isinstance(ver_major, int) or not isinstance(ver_minor, int):
        raise ProjWrongTypeError("tool.uvoxen.version.{major,minor}", "integer")  # noqa: EM101

    del proj_tool["format"]
    try:
        match ver_major, ver_minor:
            case 0, 1:
                return _pyproject_loader_v0dot1().decode(proj_tool).upgrade_to_latest()

            case 0, _:
                return _pyproject_loader().decode(proj_tool)

            case _:
                raise ProjFormatVersionUnsupportedError(ver_major, ver_minor)
    except ValueError as err:
        raise ProjDeserializeError(err) from err


def pyproject(cfg: defs.Config) -> Project:
    """Parse the pyproject.toml configuration."""
    cfg.log.debug("Parsing the pyproject.toml file")
    with pathlib.Path("pyproject.toml").open(mode="rb") as proj_file:
        proj: Final = tomllib.load(proj_file)
    try:
        proj_tool: Final = proj["tool"]["uvoxen"]
    except (KeyError, TypeError) as err:
        raise ProjMissingError("tool.uvoxen") from err  # noqa: EM101

    parsed: Final = _pyproject_load(proj_tool)

    try:
        classifiers: Final = proj["project"]["classifiers"]
    except (KeyError, TypeError) as err:
        raise ProjMissingError("project.classifiers") from err  # noqa: EM101
    if not isinstance(classifiers, list):
        raise ProjWrongTypeError("project.classifiers", "list")  # noqa: EM101
    python_supported: Final = [
        found.group("version")
        for found in (RE_PY_SUPPORTED.match(line) for line in classifiers)
        if found
    ]

    dep_groups: Final = proj.get("dependency-groups")
    match dep_groups:
        case None:
            python_dependency_groups = {}

        case dict(groups):
            python_dependency_groups = {
                name: _parse_dep_group(name, dep_group) for (name, dep_group) in groups.items()
            }

        case _:
            raise ProjWrongTypeError("dependency-groups", "table")  # noqa: EM101

    return Project(
        build_project=parsed.build_project,
        defs=parsed.defs,
        env=parsed.env,
        envlist=parsed.envlist,
        python_dependency_groups=python_dependency_groups,
        python_supported=python_supported,
        req=parsed.req,
        tox=parsed.tox,
    )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Common definitions for the uvoxen library."""

from __future__ import annotations

import dataclasses
import typing


if typing.TYPE_CHECKING:
    import logging
    from typing import Final


VERSION: Final = "0.1.1"
"""The uvoxen library version, semver-like."""


FEATURES: Final = {
    "uvoxen": VERSION,
    "format-min": "0.1",
    "format-current": "0.2",
    "req-generate": "0.1",
    "tox-expand-vars": "0.1",
    "tox-generate": "0.2",
    "tox-run": "0.1",
    "uv-generate": "0.1",
    "uv-run": "0.2",
}
"""The list of features supported by the uvoxen library."""


@dataclasses.dataclass
class Error(Exception):
    """An error that occurred while processing the test settings."""

    def __str__(self) -> str:
        """Provide a human-readable error message."""
        return f"uvoxen error: {self!r}"


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the uvoxen library."""

    log: logging.Logger
    """The logger to send diagnostic, informational, and error messages to."""

    python: str | None
    """The Python version to use."""

    verbose: bool
    """Verbose operation; display diagnostic output."""



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Make sure that `uvoxen run` starts up at least."""

from __future__ import annotations

import subprocess  # noqa: S404
import sys
import typing


if typing.TYPE_CHECKING:
    from typing import Final


def test_list_envs() -> None:
    """Make sure that `uvoxen list --help` and `uvoxen list` work."""
    output_help: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "list", "envs", "--help"],
        encoding="UTF-8",
    )
    assert "Usage: " in output_help
    assert "format" not in output_help.splitlines()

    lines_list: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "list", "envs"],
        encoding="UTF-8",
    ).splitlines()
    assert lines_list == [
        "docs",
        "format",
        "mypy",
        "pyupgrade",
        "reformat",
        "reuse",
        "ruff",
        "unit-tests-pytest-8",
        "uvoxen-sync-check",
        "var-expand-test",
    ]

    pythons_list: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "list", "pythons"],
        encoding="UTF-8",
    ).splitlines()
    assert pythons_list == ["3.11", "3.12", "3.13"]

    reqs_list: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "list", "reqs"],
        encoding="UTF-8",
    ).splitlines()
    assert reqs_list == [
        "docs",
        "testenv-mypy",
        "testenv-pyupgrade",
        "testenv-reuse",
        "testenv-ruff",
        "testenv-unit-tests-common",
        "testenv-unit-tests-pytest-8",
    ]


def test_uv_run_noop() -> None:
    """Make sure that `uvoxen uv run --help` and `uvoxen uv run --noop` work."""
    output_help: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "uv", "run", "--help"],
        encoding="UTF-8",
    )
    assert "Usage: " in output_help
    assert "uv sync" not in output_help.splitlines()

    lines_list: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "-v", "uv", "run", "--noop", "-e", "format"],
        encoding="UTF-8",
    ).splitlines()
    assert lines_list == [
        "rm -rf -- .venv",
        "uv sync --no-install-project --group testenv-ruff",
        (
            "uv run --no-sync -- "
            "ruff check --config ruff-base.toml --select=D,I --diff -- src/uvoxen tests/unit"
        ),
        (
            "uv run --no-sync -- "
            "ruff format --config ruff-base.toml --check --diff -- src/uvoxen tests/unit"
        ),
    ]


def test_tox_generate() -> None:
    """Make sure that `uvoxen tox generate -o -` outputs something sensible."""
    output_help: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "tox", "generate", "--help"],
        encoding="UTF-8",
    )
    assert "Usage: " in output_help
    assert "\nenvlist =\n" not in output_help.splitlines()

    output_gen: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "-v", "tox", "generate", "-o", "-"],
        encoding="UTF-8",
    )
    assert (
        """[tox]
minversion = 4.22
envlist =
  ruff
  format
  reuse
  mypy
  unit-tests-pytest-8
"""
    ) in output_gen

    assert (
        """[testenv:docs]
skip_install = True
tags =
  docs
dependency_groups =
  docs
commands =
  mkdocs build
"""
        in output_gen
    )


def test_tox_run() -> None:
    """Make sure `uvoxen tox run --noop` outputs something sensible."""
    output_help: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "tox", "run", "--help"],
        encoding="UTF-8",
    )
    assert "Usage: " in output_help
    assert "tox run-parallel" not in output_help

    lines_run: Final = subprocess.check_output(  # noqa: S603
        [
            sys.executable,
            "-m",
            "uvoxen",
            "-v",
            "tox",
            "run",
            "--noop",
            "--",
            "--some",
            "arg uments",
        ],
        encoding="UTF-8",
    ).splitlines()
    assert lines_run == [
        "uvx --with tox-uv -- tox run-parallel -- --some 'arg uments'",
    ]

    lines_run_supported: Final = subprocess.check_output(  # noqa: S603
        [sys.executable, "-m", "uvoxen", "-v", "-p", "supported", "tox", "run", "--noop"],
        encoding="UTF-8",
    ).splitlines()
    assert lines_run_supported == [
        "uvx -p 3.11 --with tox-uv -- tox run-parallel --",
        "uvx -p 3.12 --with tox-uv -- tox run-parallel --",
        "uvx -p 3.13 --with tox-uv -- tox run-parallel --",
    ]
